//****************************************************
//***** FRISK OLED Gray Scale Test               *****
//*************************** By Minarai Shokunin ****
//*************************** 2013/10/29          ****
//****************************************************

#include "p24FJ64GA002.h"
#include "oled.h"

extern unsigned int GSVirtualScreen[8][128];
extern unsigned char areasrc[];

// ************ prototypes ***************************
extern void OLED_init();
extern void WriteData_GS(int,int);
extern void Copy2OLED_GS(void);
extern void GR_FILL_GS(unsigned char);
extern void GR_PSET_GS(int,int,unsigned char);
extern void GR_LINE_GS(int,int,int,int,unsigned char);
extern void GR_PUTCH_GS(int,int,unsigned char, unsigned char);
extern void GR_PUTNUM_GS(int,int,int, unsigned char);
extern void GR_PUTSTR_GS(int,int,unsigned char *, unsigned char);
extern void GR_PUT8x8(int,int,int);

void game_init(void);
void round_init(void);
void disp_area(int);
void move_enemies(void); 
void change_enemy_dir(int);
void game_over(void);

void wait_some(int);
void wait_long(int t);
unsigned int get_adc(int);
void get_xy(void);
void disp_opening(void);
void clear_demo(void);
void lost_demo(void);

// ************ CONFIGRATION BITS ********************
_CONFIG1(JTAGEN_OFF & GCP_OFF & GWRP_OFF & BKBUG_OFF & COE_OFF & ICS_PGx1 & FWDTEN_OFF )
_CONFIG2(IESO_OFF & FNOSC_FRCPLL & FCKSM_CSDCMD & OSCIOFNC_OFF & IOL1WAY_OFF & I2C1SEL_PRI & POSCMOD_NONE)


// ************ sound data ***************************
int sound_freq[] = {
	// C  , C# , D  , D# , E  , F  , F# , G  , G# , A  , A# , B 	
	  130 ,139 ,147 ,156 ,165 ,175 ,185 ,196 ,208 ,220 ,233 ,247,
	  261 ,277 ,294 ,311 ,330 ,350 ,370 ,392 ,415 ,440 ,466 ,494,
      523, 554 ,587 ,622 ,659 ,698 ,740 ,783 ,830 ,880 ,932 ,988,
      1046,1108,1175,1245,1319,1396,1480,1568,1661,1760,1865,1796,
      1,1,1,1,1,1,1,1,1,1,1,1
};
int sound_len[] = {
	2400,1200,600,300,150,		// 1, 1/2 , 1/4 , 1/8 , 1/16
    3600,1800,900,450,225,		// With Dot
	1600,800,400,200,100		// 3 renpu
};

const int sound_test[] = {
	24,1,26,1,28,1,29,1,31,1,33,1,35,1,
	0xffff,0xffff
};

const int sound_pi0[] = {
	36,4,40,4,
	0xffff,0xffff
};

const int sound_pi1[] = {
	24,4,28,4,31,3,
	0xffff,0xffff
};
const int sound_pi2[] = {
	24,4,28,4,31,4,28,4,31,3,
	0xffff,0xffff
};

const int sound_clear[] = {
	24,4,28,4,31,4,24,4,28,4,31,4,24,4,28,4,31,4,36,2,
	0xffff,0xffff
};

const int sound_start[] = {
	24,3,31,3,29,3,28,3,31,3,29,3,31,2,
	26,3,33,3,31,3,28,3,33,3,31,3,33,2,
	24,3,31,3,29,3,28,3,31,3,29,3,31,3,35,3,36,2,	
	0xffff,0xffff
};


const int sound_lost[] = {
	35,3,33,3,31,3,29,3,28,3,26,3,24,2,
	0xffff,0xffff
};

const int sound_gameover[] = {
	28,2,24,2,28,2,24,2,28,2,24,2,28,2,24,1,
	0xffff,0xffff
};

// ********** Enemy direction decidion table ***********
unsigned char decide_table[] = {
	// **** Enemy type 0 ******
	2,1,0,3,			// L D R U
	0,1,2,3,			// R D L U
	0,3,2,1,			// R U L D
	2,3,0,1,			// L U R D
	// **** Enemy type 1 ******
	1,2,3,0,			// D L U R
	1,0,3,2,			// D R U L
	3,0,1,2,			// U R D L
	3,2,1,0,			// U L D R
	// **** Enemy type 0 ******
	2,1,0,3,			// L D R U
	0,1,2,3,			// R D L U
	0,3,2,1,			// R U L D
	2,3,0,1,			// L U R D
	// **** Enemy type 1 ******
	1,2,3,0,			// D L U R
	1,0,3,2,			// D R U L
	3,0,1,2,			// U R D L
	3,2,1,0,			// U L D R
};
// ********** global variables *******

unsigned char pac_area[9][15];
int pac_x,pac_y,pac_dir,pac_left;
int dot_left,poweruptimer;
int enemy_x[4];
int enemy_y[4];
int enemy_dir[4];
int enemy_param[4];
int score,hiscore;
int fexit,pac_hit;

int sound_enabled;			// Sound enable flag
int sound_left;				// left count for current tone
int sound_tone;				// current tone
int *sound_pointer;			// current tone/len data

#define KEY_ON 		1
#define KEY_OFF 	0
#define KEY_PUSHED	2
#define KEY_L 		0
#define KEY_R 		1
#define KEY_S 		2

int key_status[3];
int joy_x;
int joy_y;
int joy_sw;


// *********************************************
// ************ MAIN ***************************
// *********************************************

int main(void)
{
	int i;

	// ************ CPU Initialize **********

	_RCDIV = 0;							// PostScaler 32MHz
	TRISB = 0b0000011100000011;			// PORTB 15-11,7-0 = output  ; 10-9 = input ; 1-0 = input for analog
	CNPU1 = 0xffff;						// Set pull-ups
	CNPU2 = 0xffff;						// Set pull-ups

	// *********** Sound Itinialize ***************

	sound_enabled = NO;
	sound_left = 0;

	RPOR3bits.RP6R 		= 18;   // RP6 = OSC1
    PR2             	= 32000; 	// Timer2 period 
    T2CONbits.TCKPS 	= 0;    // Timer2 Pre scaler 1:1
    T2CONbits.TON   	= 0;    // Timer2 OFF
    OC1CONbits.OCTSEL 	= 0; 
    OC1CONbits.OCM    	= 5;      
    OC1R              	= PR2;   
    OC1RS             	= PR2/2; 

	// *********** ADC initialize *******************
	CMCON 				= 0;		
	AD1PCFG				= 0xFFF3;		// Use AN2,AN3 as analog
	AD1CON1 			= 0x2008;
	AD1CON2 			= 0x0000;
	AD1CON3 			= 0x0000;
	AD1CHS  			= 0x0000;
	AD1CSSL 			= 0x0000;



	// *********** Display Itinialize ***************

	OLED_init();
	wait_some(1000);
	GR_FILL_GS(0);

	// *********** sound init ***************

 	PR1 = 4000; 
	sound_enabled = NO;
    T1CONbits.TCKPS 	= 2;    // Timer1 Pre scaler 1:64 = 16ms
    T1CONbits.TON   	= 1;    // Timer1 ON
 	IEC0bits.T1IE = 1;  // Timer1 interrupt enable
	sound_pointer = sound_pi0;
	sound_enabled = NO;

	// ***********  MAIN LOOP **************
	hiscore = 0;
	while (1) {
		disp_opening();
		pac_left =3;
		pac_hit = NO;
		round_init();
		sound_pointer = sound_start;
		sound_enabled = YES;
		GR_FILL_GS(0);
		disp_area(pac_y);
		while(sound_enabled==YES){;}
		wait_long(20);
		score=0;

		// *** Loop while PAC man exits ***

		while(pac_left >0) {
			round_init();
			fexit=NO;
			pac_hit = NO;

			// **** Loop until clear or lost ****

			while(fexit==NO) {
				GR_FILL_GS(0);
				move_packman();
				move_enemies();
				disp_area(pac_y);
				wait_long(4);
				if (dot_left==0) {fexit=YES;clear_demo();}
				if (pac_hit==YES) {fexit=YES;lost_demo();pac_left--;}
			}	
		}
		game_over();
	}
}


// *********************************************
// ************ AP Sub routines    **********
// *********************************************
void game_over(void) {
	sound_pointer = sound_gameover;
	sound_enabled = YES;
	while(sound_enabled==YES){;}
	GR_FILL_GS(0);
	GR_LINE_GS(0,0,127,0,3);
	GR_LINE_GS(0,63,127,63,3);
	GR_LINE_GS(0,0,0,63,3);
	GR_LINE_GS(127,0,127,63,3);
	GR_PUTSTR_GS(10,8,"GAME OVER",2);
	GR_PUTSTR_GS(10,24,"SCORE:",2);
	GR_PUTNUM_GS(70,24,score,3);
	GR_PUTSTR_GS(10,48,"PUSH L",3);
	while (SW_L == 1) {
	}	
	while (SW_L == 0) {
	}	
}

void disp_opening(void) {
	GR_FILL_GS(0);
	GR_LINE_GS(0,0,127,0,3);
	GR_LINE_GS(0,63,127,63,3);
	GR_LINE_GS(0,0,0,63,3);
	GR_LINE_GS(127,0,127,63,3);
	GR_PUTSTR_GS(10,8,"MINI",2);
	GR_PUTSTR_GS(10,16,"PACMAN",2);
	GR_PUT8x8(20,32,0);
	GR_PUT8x8(30,32,0);
	GR_PUT8x8(40,32,0);
	GR_PUT8x8(60,32,9);
	GR_PUTSTR_GS(10,48,"PUSH L",3);
	while (SW_L == 1) {
	}	
	while (SW_L == 0) {
	}	

}
void clear_demo(void) {
	sound_pointer = sound_clear;
	sound_enabled = YES;
	while(sound_enabled==YES){;}
}
void lost_demo(void) {
	sound_pointer = sound_lost;
	sound_enabled = YES;
	while(sound_enabled==YES){;}
}


void move_enemies(void) {
	int i;
	int enemy_x0,enemy_y0;

	for (i=0;i<4;i++) {
		if (enemy_dir[i] ==4) {
				if (enemy_x[i]>4) {enemy_x[i]--;}
				if (enemy_x[i]<4) {enemy_x[i]++;}
				if (enemy_y[i]>7) {enemy_y[i]--;}
				if (enemy_y[i]<7) {enemy_y[i]++;}
				if ((enemy_x[i]==4)&&(enemy_y[i]==7)) {
					enemy_dir[i] = 3;
				}
		}else{
			if ((rand()&3)==0) {change_enemy_dir(i);}
			enemy_x0= enemy_x[i];enemy_y0=enemy_y[i];
			switch (enemy_dir[i]) {
				case 0:
						if ((pac_area[enemy_x0][enemy_y0]&1) == 0) {
							enemy_x0++;
						}else{
							change_enemy_dir(i);
						}
						break;
				case 1:
						if ((pac_area[enemy_x0][enemy_y0]&2) == 0) {
							enemy_y0++;
						}else{
							change_enemy_dir(i);
						}
						break;
				case 2:
						if ((pac_area[enemy_x0][enemy_y0]&4) == 0) {
							enemy_x0--;
						}else{
							change_enemy_dir(i);
						}
						break;
				case 3:
						if ((pac_area[enemy_x0][enemy_y0]&8) == 0) {
							enemy_y0--;
						}else{
							change_enemy_dir(i);
						}
						break;
				deafult:	
						break;
			}
			enemy_x[i] = enemy_x0; enemy_y[i] = enemy_y0;
			if ((enemy_x[i]==0)&&(enemy_y[i]==7)){ 
					enemy_x[i]=8;enemy_dir[i]=2;
			}else{
				if ((enemy_x[i]==8)&&(enemy_y[i]==7)){ 
					enemy_x[i]=0;enemy_dir[i]=0;
				}
			}
		}
	}
}
void change_enemy_dir(int no) {
	int dir,i,j,v,x0,y0,typ;
	unsigned char c0;
	dir = 0; i=0;
	v=0;
	typ = rand()&0x03*16;
	if (pac_x < enemy_x[no]) {dir += 4;} 
	if (pac_y < enemy_y[no]) {dir += 8;}
	c0 = pac_area[enemy_x[no]][enemy_y[no]]; 
	while (v == 0) {
		if (poweruptimer>0) {j=3-i;}else{j=i;}
		switch (decide_table[typ+dir+j]) {
			case 0:
				if ((c0&1) == 0) {v=1;}
				break;
			case 1:
				if ((c0&2) == 0) {v=1;}
				break;
			case 2:
				if ((c0&4) == 0) {v=1;}
				break;
			case 3:
				if ((c0&8) == 0) {v=1;}
				break;
			default:
				v=1;i=3;
				break;
		}
		i++;
	}
	enemy_dir[no] = decide_table[typ+dir+j];
}

void move_packman(void) {
	unsigned char c0;
	int moved,i;

	get_xy();
	GR_PUTSTR_GS(80,8,"HI",2);
	GR_PUTNUM_GS(80,16,hiscore,3);
	GR_PUTSTR_GS(80+32,16,"0",3);
	GR_PUTSTR_GS(80,24,"SCR",2);
	GR_PUTNUM_GS(80,32,score,3);
	GR_PUTSTR_GS(80+32,32,"0",3);
	for (i=0;i<pac_left;i++) { 		GR_PUT8x8(80+i*10,48,0x09); 	}  	c0 = pac_area[pac_x][pac_y];  	// ***** Get Power up item ************ 	if ((c0 & 0x20) !=0) { 		score += 2; 		poweruptimer = 30; 		pac_area[pac_x][pac_y] &= 0x0f; 		dot_left--; 		sound_pointer = sound_pi1; 		sound_enabled=YES; 	} 	// ***** Get dot *********************** 	if ((c0 & 0x10) !=0) {      // Dot counter 		score += 1; 		pac_area[pac_x][pac_y] &= 0x0f; 		dot_left--; 		if (sound_enabled==NO) { 			sound_pointer = sound_pi0; 			sound_enabled=YES; 		} 	}  	moved=0; 	if ((joy_x>75)&&(pac_x<=8)&&((c0&1)==0)) {
			if ((pac_x==8)&&(pac_y==7)) {
				pac_x=0;
			}else{
				pac_x++;
			}
			pac_dir=0;moved=1;
	}
	if ((joy_x<30)&&(pac_x>=0)&&((c0&4)==0)) {
			if ((pac_x==0)&&(pac_y==7)) {
				pac_x=8;
			}else{
				pac_x--;
			}
			pac_dir=1;moved=1;
	}
	if (moved==0) {
		if ((joy_y>75)&&(pac_y<14)&&((c0&2)==0)) {pac_y++;pac_dir=3;}
		if ((joy_y<30)&&(pac_y>0)&&((c0&8)==0)) {pac_y--;pac_dir=2;}
	}

	if (poweruptimer >0) {
		poweruptimer--;
	}

	for (i=0;i<4;i++) {
		if ((enemy_x[i]==pac_x)&&(enemy_y[i]==pac_y)&&(enemy_dir[i]<4)) {
			if (poweruptimer >0) {
				score += 5;
				enemy_dir[i]=4;
				sound_pointer = sound_pi2;
				sound_enabled = YES;
			}else{
				pac_hit = YES;
			}
		}
	}
}


void disp_area(int y0) {
	static int cnt;
	int x,y,yl,i,flag;
	int c0,c1;
	yl=0;
	for (y=(y0-3);y<(y0+5);y++) {
		if ((y>=0)&&(y<15)) {
			for (x=0;x<9;x++) {
				c0 = pac_area[x][y];c1 = c0 & 0x0f;
				if (c1>0) {
					GR_PUT8x8(x*8,yl*8,c1+18);
				}
				//******* ENEMIES *********
				flag =0;
				for (i=0;i<4;i++) {
					if ((enemy_x[i]==pac_x)&&(enemy_y[i]==pac_y)) {
							if (poweruptimer >0) {
								score += 5;
								enemy_dir[i]=4;
								sound_pointer = sound_pi2;
								sound_enabled = YES;
							}else{
								pac_hit = YES;
							}
					}else{
						if ((enemy_x[i]==x)&&(enemy_y[i]==y)) {
							if ((poweruptimer >0)||(enemy_dir[i]==4)) {
								GR_PUT8x8(x*8,yl*8,4);
							}else{
								GR_PUT8x8(x*8,yl*8,enemy_dir[i]);
							}
							flag = 1;
						}
					}
				}
				if (flag==0) {
					//******* DOTS *********	
					if ((c0 & 0x20)==0x20) {
						GR_PUT8x8(x*8,yl*8,18);
					}
					if ((c0 & 0x10)==0x10) {
						GR_PUT8x8(x*8,yl*8,17);
					}
				}
			}
		}
		yl++;
	}
	//******* PACMAN **********
	GR_PUT8x8(pac_x*8,3*8,5+4*cnt+pac_dir);
	cnt++;	if (cnt>2) {cnt=0;}
}

void round_init(void) {
	int x,y,i;

	for (i=0;i<4;i++) {
		enemy_x[i]=4;
		enemy_y[i]=7;
		enemy_dir[i]=3;
		enemy_param[i]=3;
	}

	poweruptimer =0;

	if (pac_hit == NO) {
		dot_left = 96;
		for (y=0;y<15;y++) {
			for (x=0;x<9;x++) {
				pac_area[x][y] = areasrc[y*9+x];
			}
		}
	}
	pac_x = 4;pac_y=8;
}

// *********************************************
// ************ BASIC Sub routines    **********
// *********************************************
void get_xy(void) {

	joy_x = get_adc(3)/10;
	joy_y = get_adc(2)/10;

}

unsigned int get_adc(int ch) {
	int ret,t;

	AD1CHS = ch;
	AD1CON1bits.ADON = 1;		// ADC Start
	AD1CON1bits.SAMP = 1;		// Sampling start
	for (t=0;t<5;t++) {};
	AD1CON1bits.SAMP = 0;
	while(!AD1CON1bits.DONE) {};
	ret = ADC1BUF0;
	return ret;
}

//********** Simple wait loop *************
void wait_some(int t)
{
	int i,j;
	for (i=0;i